# RTLS Monitor

This application connects to a running `websocket_server` agent process connected to RTLS devices connected via serial port.

## Workaround

An issue exists in the rtls_agent_cli binary that is included with CC13x2/26x2 SDK 3.20.
An updated binary is hosted on the [RTLS E2E Known Issues Page](http://e2e.ti.com/support/wireless-connectivity/bluetooth/f/538/t/770218).
Please download this binary and use it in place of the one hosted in the SDK.
For the rest of this document references to `rtls_agent_cli` refer to the patched version.

For those using Python directly, the fix is also posted.

## Prerequisites

- CC13x2/26x2 SDK 3.20
  - [View in Resource Explorer](http://dev.ti.com/tirex/explore/node?node=AOSQDVXMohlV5LElLx5wxA__pTTHBmu__LATEST)
  - [Download installer](http://www.ti.com/tool/SIMPLELINK-CC13X2-26X2-SDK)
- CC26x2 LaunchPads programmed with example apps
  - rtls_master (one device)
  - rtls_slave (one device)
  - rtls_passive (one or more devices)
- Running RTLS Agent (`websocket_server`)
  - `<sdk folder>/tools/ble5stack/rtls_agent/agent/rtls_agent_cli.py`
  - **NOTE:** Agent tool should not connect to `rtls_slave` as it has no serial communication.

Please see `README.md` in the `tools/ble5stack/rtls_agent/` folder of the SDK for information on how to run the agent.

Run `rtls_agent_cli`, press 'a' to auto detect the devices. Then press enter to auto detect and connect to
`rtls_master` and `rtls_passive`.

## Using

If you launch `rtls_agent_cli` with no command line arguments, it will default to opening a websocket on port `8766`, which is the default configuration in the GUI.

Once the agent software is running, select if automatic RTLS should be enabled (checkbox next to reload button) and click `Connect`.
The behavior of automatic RTLS is described below.

## Selecting capabilities

By default, both AoA and ToF will be enabled.
However, the user can disable features

In order to do AoA, an [AoA BoosterPack](http://www.ti.com/tool/BOOSTXL-AOA) must
be attached.

If an AoA BoosterPack is not mounted, then the AoA results should be disregarded and are not valid.
This is because no antenna switching occurs.

If you wish to prevent AoA from being initiated by the GUI, then it should be removed
from the embedded software's reported capabilities. These are populated in
`main.c` of the `rtls_*` samples applications.

For example, from `rtls_passive`:

```c
    rtlsConfig.rtlsCapab = (rtlsCapabilities_e)(RTLS_CAP_RTLS_MASTER | RTLS_CAP_TOF_MASTER | RTLS_CAP_AOA_RX);
```

Remove the field `RTLS_CAP_AOA_RX` to prevent AoA.

## Automatic RTLS

If automatic RTLS is enabled, then the following steps will be performed automatically when clicking `Connect`.

 * Scan for RTLS slave devices
 * Connect to the first RTLS slave device that is detected
 * Pair and open L2CAP CoC
 * Enable AoA or ToF

If automatic RTLS is disabled, then the following must be done manually

 * Click the `Scan for Slave Devices` --> This will start a scan and put the Bluetooth address of found peers in the Remote devices list
 * Select the `rtls_slave` by clicking on the Bluetooth address in the remote devices list, then select `Connect to slave
 * Select `Enable RTLS` to start AoA or ToF measurements.

Note that the buttons will only be enabled when they can be pressed.
For example `Connect to slave` is only enabled after `RTLS_CMD_SCAN_STOP` is received.

## Modifying

Import this GUI Composer project into GUI Composer via the gallery on dev.ti.com/gallery and change either the GUI or the code (app.js).

## Further resources

More information can be found at the following links

 * [BLE5-Stack User's Guide RTLS Chapter](http://dev.ti.com/tirex/explore/node?node=AOimuSWjap.4RuDbcp7OqA__pTTHBmu__LATEST)
 * [RTLS SimpleLink Academy](http://dev.ti.com/tirex/explore/node?node=AHjJ8SNDuLXt3h6qHE4-OA__pTTHBmu__LATEST)