/*****************************************************************
 * Copyright (c) 2017 Texas Instruments and others
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Paul Gingrich - Initial API and implementation
 *****************************************************************/
var gc = gc || {};
gc.databind = gc.databind || {};
gc.databind.internal = gc.databind.internal || {};
gc.databind.internal.reg = gc.databind.internal.reg || {};

(function()
{
    var reducerMaxIndexDifferent = function(otherArray, maxIndex, cachedVal, index, arr) {
        return (cachedVal !== otherArray[index]) ? index : maxIndex;
    };
    var nullProgressCounter = new gc.databind.IProgressCounter();

    /**
     * Callback used by writeValue.
     *
     * @callback gc.databind.AbstractAsyncBindValue#onWriteValue
     */

    var onDone = function(that, newValueObj)
    {
        var newValue;
        var index;
        if (newValueObj) {
            newValue = newValueObj.value;
            index = newValueObj.coreIndex;
        }
        const IDLE = that.getStateConstant('IDLE');
        const READ = that.getStateConstant('READ');
        const WRITE = that.getStateConstant('WRITE');
        const DELAYED_WRITE = that.getStateConstant('DELAYED_WRITE');
        const DELAYED_READ = that.getStateConstant('DELAYED_READ');

        if (that._state === DELAYED_WRITE)
        {
            that.changeState(WRITE);
            var deferredValue;
            if (Array.isArray(that.fCachedValue)) {
                deferredValue = that.fCachedValue.slice();    // save the deferred write value.  make copy of array
                that.fCachedValue = that.fCommittedValue.slice();  // make sure we write the lastWrittenValue instead of the deferred value
                if (newValue !== undefined) {
                    if (index === -1) {
                        for (var i = 0; i < newValue.length; i++) {
                            (function (i) {
                                that.writeValueFor(i, function () {
                                    if (i === newValue.length - 1) {
                                        if (that._delayedProgress) {
                                            that._delayedProgress.done();
                                            that._delayedProgress = null;
                                        }
                                        onDone(that);
                                    }
                                });
                            })(i);
                        }
                    } else {
                        if (index !== undefined) {
                            that.writeValuerFor(index,function() {
                                if (that._delayedProgress) {
                                    that._delayedProgress.done();
                                    that._delayedProgress = null;
                                    onDone(that);
                                }
                            });
                        }
                    }
                }
            } else {
                deferredValue = that.fCachedValue;    // save the deferred write value
                that.fCachedValue = that.fCommittedValue;  // make sure we write the lastWrittenValue instead of the deferred value
                that.writeValue(function()
                {
                    if (that._delayedProgress)
                    {
                        that._delayedProgress.done();
                        that._delayedProgress = null;
                    }
                    onDone(that);
                });
            }
            that.fCachedValue = deferredValue;      // restore deferred value
        }
        else if (that._state === WRITE || that._state === DELAYED_READ || (that._state === READ && that.isConnected() === false))
        {
            if (that._readable && !that.isDeferredWritePending())
            {
                // read back the value to see if the value we wrote stuck.
                that.changeState(READ);
                if ((newValue !== undefined) && Array.isArray(newValue)) {
                    var regValue = [];
                    for (var index=0; index < newValue.length; index++) {
                        (function(index) {
                            that.readValueFor(index, function (rdValue) {
                                regValue.push(rdValue);
                                if (index === newValue.length - 1) {
                                    if (that._delayedProgress) {
                                        that._delayedProgress.done();
                                        that._delayedProgress = null;
                                    }
                                    onDone(that, {value: regValue, coreIndex: index});
                                }
                            });
                        })(index);
                    }
                } else {
                    that.readValue(function (rdValue) {
                        if (that._delayedProgress) {
                            that._delayedProgress.done();
                            that._delayedProgress = null;
                        }
                        onDone(that, {value: rdValue, coreIndex: index});
                    });
                }
            }
            else
            {
                that.changeState(IDLE);
            }
        }
        else if (that._state === READ)
        {
            that.changeState(IDLE, newValue);
            // ignore newly read value if a deferred value is pending to be written
            if (!that.isDeferredWritePending())
            {
                if (Array.isArray(newValue)) {
                    that.fCommittedValue = newValue.slice();
                } else {
                    if (!Array.isArray(that.fCommittedValue)) {
                        that.fCommittedValue = [];
                    }
                    while (index >= that.fCommittedValue.length) {
                        that.fCommittedValue.push(undefined);
                    }
                    that.fCommittedValue[index] = newValue;
                }
                that.updateValue(that.fCommittedValue);
                that.setStale(false);
            }
        }
    };

    gc.databind.internal.reg.RegisterAllBind = function(uri, model, refreshIntervalProvider, symbolData)
    {
        try {
            gc.databind.AbstractAsyncBindValue.call(this, 'array');

            var that = this;
            that.uri = uri;
            that.parentModel = model;
            that.activeIndex = 0;
            that.fCachedValue = [];
            that.fCommittedValue = [];

            that._onFailure = function (errMsg) {
                if (that.parentModel.isConnected()) {
                    that.reportErrorStatus(errMsg);
                }
                var callback = that._callback;
                that._callback = undefined;
                if (callback) {
                    callback(that.fCachedValue);  // don't record a new value, keep the same value as before.
                }
            };

            that._onSuccess = function (index, result) {
                // clear errors on successful read
                that.reportCriticalError(null);

                if ((index === -1) && (!Array.isArray(result))) {
                    result = [result];
                }
                return {coreIndex: index, value: result};
            };

            // add refresh listener and create dispose() method to remove listener when done.
            refreshIntervalProvider.addRefreshListener(that);
            that.dispose = function () {
                refreshIntervalProvider.removeRefreshListener(that);
            };

            this.updateRegisterInfo(symbolData);

        } catch(ex) {
            console.log("RegisterAllBind: exception = "+ex);
        }

    };


    gc.databind.internal.reg.RegisterAllBind.prototype = new gc.databind.AbstractAsyncBindValue('array');

    gc.databind.internal.reg.RegisterAllBind.prototype.getValueFor = function(index) {
        var result = undefined;
        if (index >= 0) {
            if (this.fCachedValue && Array.isArray(this.fCachedValue) && this.fCachedValue.length > index) {
                result = this.fCachedValue[index];
            }
        } else {
            result = this.fCachedValue;
        }
        return result;
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.setValueFor = function(newValue, index, forceWrite) {
        if (this.isReadOnly()) {
            return;
        }
        if (index >= 0) {
            while (index >= this.fCachedValue.length) {
                this.fCachedValue.push(this.defaultValue);
            }
        }

        var oldValue = this.fCachedValue[index];
        this.fCachedValue[index] = newValue;
        if (!this.fDeferredMode) {
            if ((this.uri.indexOf("$cores.all.") === 0) && (forceWrite || ((newValue !== undefined) && (newValue !== oldValue)))) {
                var regUri = this.uri.substring("$cores.all.".length);
                this.parentModel.writeValue(regUri, this.fCachedValue[index], index).then(function () {
                    this.notifyChangedListeners(oldValue, newValue, null);
                }.bind(this)).fail(this._onFailure);
            }
        }
        return newValue;
    };
    // same as setValue but it doesn't commit the value to the target
    gc.databind.internal.reg.RegisterAllBind.prototype.updateValueFor = function(value, index, progress, skipStreamingListeners) {
        if ((index >= 0) &&  (index <= this.fCachedValue.length)) {
            if (value !== this.fCachedValue[index]) {
                // update the entire array to ensure that AbstractBindValue.deepNotEquals will notice the change and notify listeners
                var newArray = this.fCachedValue.slice();
                newArray[index] = value;
                this.updateValue(newArray, progress, skipStreamingListeners);
            }
        }
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.setActiveIndex = function(newIndex) {
        this.activeIndex = newIndex;
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.getActiveIndex = function() {
        return this.activeIndex;
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.getValueCommittedFor = function(coreIndex){
        if (this.fCommittedValue && Array.isArray(this.fCommittedValue) && (coreIndex < this.fCommittedValue.length)) {
            return this.fCommittedValue[coreIndex];
        } else {
            // initialize the fCommittedValue array so that it includes an entry for coreIndex
            this.setValueCommittedFor(coreIndex, undefined);
            return undefined;
        }
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.setValueCommittedFor = function(coreIndex, value) {
        if (!(this.fCommittedValue && Array.isArray(this.fCommittedValue))) {
            this.fCommittedValue = [];
        }
        if (coreIndex >= 0) {
            while (coreIndex >= this.fCommittedValue.length) {
                this.fCommittedValue.push(undefined);
            }
            this.fCommittedValue[coreIndex] = value;
        }
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.clearDeferredWriteFor = function(coreIndex) {
        if (coreIndex >= 0 &&
            (Array.isArray(this.fCommittedValue) && (coreIndex < this.fCommittedValue.length)) &&
            (Array.isArray(this.fCachedValue) && (coreIndex < this.fCachedValue.length))) {
            this.fCommittedValue[coreIndex] = this.fCachedValue[coreIndex];
        }
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.setDeferredMode = function(deferredMode, forceWrite) {
        deferredMode = deferredMode || false;
        if (deferredMode !== this.fDeferredMode) {

            var oldValue = this.fCommittedValue.slice();
            var newValue = this.fCachedValue.slice();

            gc.databind.AbstractLookupBindValue.prototype.setDeferredMode.call(this, deferredMode);

            // write deferred value when transitioning out of deferred mode, if appropriate.
            if (!deferredMode) {
                var isValueChanged = false;
                if (oldValue.length !== newValue.length) {
                    isValueChanged = true;
                } else {
                    isValueChanged = (oldValue.reduce(reducerMaxIndexDifferent.bind(this, newValue), -1) >= 0);
                }
                if (forceWrite || isValueChanged) {
                    this.onValueChanged(oldValue, newValue, null);
                }
            } else {
                this.fCommittedValue = newValue;
            }
        }
    };
    gc.databind.internal.reg.RegisterAllBind.prototype.onValueChanged = function(oldValue, newValue, progress)
    {
        var that = this;
        that.setStale(false); // ensure no longer stale if value set first
        // before reading from target.

        if (this._writable && !that.fDeferredMode)
        {
            if (isDeviceArray) {
                if (Array.isArray(this.fCachedValue)) {
                    this.fCommittedValue = this.fCachedValue.slice();
                } else {
                    this.fCommittedValue = this.fCachedValue;
                }
                if (that._state === IDLE || that._state === ERROR_STATE)
                {
                    that.changeState(WRITE);
                    progress = progress || nullProgressCounter;
                    progress.wait();
                    if ((isDeviceArray && Array.isArray(this.fCachedValue))) {
                        var maxIndex = -1;
                        if (oldValue.length !== newValue.length) {
                            maxIndex = 0;
                        } else {
                            maxIndex = oldValue.reduce(reducerMaxIndexDifferent.bind(this, newValue), -1);
                        }
                        if (maxIndex >= 0) {
                            for (var index=0; index <= maxIndex; index++) {
                                (function(index) {  //Immediately Invoked Function Execution. For ES6, use for (let index=0... instead.
                                    if (newValue[index] !== oldValue[index]) {

                                        // before reading from target.
                                        /* dispatch a custom event to log the write command for script script recording */
                                        var coreUri = that.uri.replace('$cores.all', '$cores.' + index);
                                        document.dispatchEvent(new CustomEvent('scripting-log-command', {
                                            detail: {
                                                command: 'write',
                                                modelId: that.parentModel.getName(),
                                                arguments: [coreUri, newValue[index]]
                                            }
                                        }));

                                        that.writeValueFor(index, function () {
                                            if (index === maxIndex) {
                                                progress.done();
                                                onDone(that, newValue);
                                            }
                                        });
                                    }
                                })(index);
                            }
                        }
                    }
                } else {
                    /* dispatch a custom event to log the write command for script script recording */
                    document.dispatchEvent(new CustomEvent('scripting-log-command', { detail: {
                            command: 'write',
                            modelId: that.parentModel.getName(),
                            arguments: [that.uri, newValue]
                        }}));
                    this.fCommittedValue = this.fCachedValue;
                    if (that._state === IDLE || that._state === ERROR_STATE)
                    {
                        that.changeState(WRITE);
                        progress.wait();
                        that.writeValue(function()
                        {
                            progress.done();
                            onDone(that);
                        });
                    }
                }
            }
            else
            {
                that.changeState(DELAYED_WRITE);
                if (that._delayedProgress)
                {
                    that._delayedProgress.done();
                }
                progress.wait();
                that._delayedProgress = progress;
            }
        }
    };


    gc.databind.internal.reg.RegisterAllBind.prototype.writeValue = function(callbackArg)
    {
        var callback = callbackArg;
        // write this register for all cores by passing in -1 as the core index
        this.parentModel.writeValue(this.uri, this.fCachedValue, -1).then(function(result) {
            callback(result);
        }.bind(this)).fail(
            this._onFailure
        );
    };
    gc.databind.internal.reg.RegisterAllBind.prototype.writeValueFor = function(index, callbackArg)
    {
        var callback = callbackArg;
        // write this register for all cores by passing in -1 as the core index
        this.parentModel.writeValue(this.uri, this.fCachedValue[index], index).then(function(result) {
            callback(result);
        }.bind(this)).fail(
            this._onFailure
        );
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.readValue = function(callbackArg)
    {
        var callback = callbackArg;
        // read this register for all cores by passing in -1 as the core index
        this.parentModel.readValue(this.uri, -1).then(function(result){
            return this._onSuccess(-1, result)
        }.bind(this)).then(function(resultObj){
            return callback(resultObj.value);
        }.bind(this)).fail(
            this._onFailure
        );
    };
    gc.databind.internal.reg.RegisterAllBind.prototype.readValueFor = function(indexArg, callbackArg)
    {
        var callback = callbackArg;
        var index = indexArg;
        // read this register for the core specified by index
        this.parentModel.readValue(this.uri, index).then(function(result){
            return this._onSuccess(index,result);
        }.bind(this)).then(function(resultObj){
            return callback(resultObj.value);
        }.bind(this)).fail(
            this._onFailure
        );
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.reportErrorStatus = function(errorMessage)
    {
        var status = null;
        if (errorMessage && errorMessage.length > 0)
        {
            status = gc.databind.AbstractStatus.createErrorStatus(errorMessage, 'target');
        }
        this.reportCriticalError(status);
    };

    gc.databind.internal.reg.RegisterAllBind.prototype.updateRegisterInfo = function(symbolData)
    {
        if (symbolData)
        {
            this.defaultValue = gc.utils.string2value(symbolData.value || symbolData['default']);
            if (this.defaultValue !== undefined)  // restore default for the new device, before reading the actual value.
            {
                var defaultValues = [];
                for (var i=0; i< this.fCachedValue.length; i++) {
                    defaultValues.push(this.defaultValue);
                }
                this.updateValue(defaultValues, undefined, true);
                this.setStale(false);
            }
            // support for qualifiers in register symbol data
            if (!this._readable)              // remove existing qualifiers
            {
                this._readable = true;
            }
            if (!this._writable)
            {
                this._writable = true;
            }
            if (!this._volatile)
            {
                this._volatile = true;
            }

            var type = symbolData.type;
            if (type)
            {
                // add any new qualifiers
                if (type === 'R')
                {
                    type = 'readonly';
                }
                else if (type === 'W')
                {
                    type = 'writeonly';
                }
                this.addQualifier(type);
            }

            // clear errors on successful read
            this.reportCriticalError(null);
        }
    };

    //----------------------------------

    gc.databind.internal.reg.RegisterActiveBind = function(uri, registerBind, activeIndexBind ) {
        // binary operator adds listeners to figure out when either the register bind or activeIndexBind changes
        // when array operator changes it will call setIndex on the registerBind.  (the first param specifies the new index)
        gc.databind.internal.expressionParser.ArrayOperator.call(this,registerBind, activeIndexBind);
        this.uri = uri;
        this.registerBind = registerBind;
        if (!this.registerBind.parentModel && registerBind.parentBind) {
            this.registerBind.parentModel = registerBind.parentBind.parentModel;
        }

        this.registerAllBind = registerBind.parentBind;  // don't really have to cache it
        this.activeIndexBind = activeIndexBind;
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype = new gc.databind.internal.expressionParser.ArrayOperator();

    gc.databind.internal.reg.RegisterActiveBind.prototype.onRefresh = function(progress){
        var that = this;
        var model = this.registerAllBind && this.registerAllBind.parentModel;
        if (!model) return;
        var isDeviceArray =  model.isDeviceArray;
        var index = this.coreIndex;
        if (isDeviceArray && this.registerAllBind && this.registerAllBind.fName && this.registerAllBind.fName.indexOf("$cores.all") === 0) {
            that = this.registerAllBind;
            index = that.activeIndex;
        } else {
            that = this.registerBind;
            index = that.coreIndex;
        }
        if (that.isRefreshable()) {
            that.changeState(that.getStateConstant('READ'));
            progress = progress || nullProgressCounter;
            progress.wait();
            if (isDeviceArray) {
                var cb = function (newValue) {
                    var cbValue = newValue;
                    if (index >= 0 && Array.isArray(that.fCachedValue)) {
                        cbValue = that.fCachedValue.slice();  // copy into a new array
                        if (!cbValue) {
                            cbValue = [];
                        }
                        cbValue[index] = newValue;
                    }
                    // for backwards compatibility:
                    if (index < 0 && !Array.isArray(newValue)) {
                        cbValue = [newValue];
                    }
                    onDone(that, {coreIndex: index, value: cbValue});
                    progress.done();
                };
                if (index < 0) {
                    that.readValue(cb);
                } else {
                    that.readValueFor(index, cb);
                }
            } else {
                that.readValue(function(newValue)
                {
                    onDone(that, newValue);
                    progress.done();
                });
            }
        }
    };
    gc.databind.internal.reg.RegisterActiveBind.prototype.onRefreshAndLog = function(progress){
        /* dispatch a custom event to log the read command for script script recording */
        document.dispatchEvent(new CustomEvent('scripting-log-command', { detail: {
                command: 'read',
                modelId: this.registerBind.parentModel.getName(),
                arguments: [this.uri]
            }}));

        this.onRefresh(progress);
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype.isRefreshable = function() {
        return this.registerBind.isRefreshable();
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype.updateRegisterInfo = function(param){
        return this.registerBind.updateRegisterInfo(param);
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype.setDeferredMode = function(deferredMode, forceWrite)
    {
        // set deferred mode on the active core only, not on the registerAllBind.
        // The user can still set deferredMode explicitly on $cores.all bindings.
        this.registerBind.setDeferredMode(deferredMode, forceWrite);
        // ensure the committed value for RegisterAllBind for the active core is the same as the committed value for that core's RegisterBind
        this.registerAllBind.setValueCommittedFor(this.activeIndexBind.getValue(), this.registerBind.getValueCommitted());
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype.clearDeferredWrite = function()
    {
        this.registerBind.clearDeferredWrite();
        this.registerAllBind.clearDeferredWriteFor(this.activeIndexBind.getValue());
    };
    gc.databind.internal.reg.RegisterActiveBind.prototype.isConnected = function()
    {
        return this.registerBind.isConnected();
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype.isReadOnly = function()
    {
        return this.registerBind.isReadOnly();
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype.isDeferredWritePending = function()
    {
        return this.registerBind.isDeferredWritePending();
    };

    gc.databind.internal.reg.RegisterActiveBind.prototype.getValueCommitted = function() {
        if (this.registerAllBind) {
            return this.registerAllBind.getValueCommittedFor(this.activeIndexBind.getValue());
        } else {
            return this.registerBind.getValueCommitted();
        }
    };

}());


